// log2.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#define OPTFENCE __asm nop

/****************************************************************************
*                                  TableTest
* Inputs:
*       UINT n: Value to look up
*       UINT dummy: ignored
* Result: UINT
*       Log2(n)
* Notes: 
*       This uses table lookup.  Because we are after speed, there is no loop,
*       just inline code
****************************************************************************/

_declspec(noinline) UINT TableTest(UINT n, UINT dummy)
    {
#define X (BYTE)-1
     const static BYTE table[256] ={X,  0,  1,  X,  2,  X,  X,  X,  // 0..7
                                    3,  X,  X,  X,  X,  X,  X,  X,  // 8..15
                                    4,  X,  X,  X,  X,  X,  X,  X,  // 16..23
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 24..31
                                    5,  X,  X,  X,  X,  X,  X,  X,  // 32..39
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 40..47
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 48..55
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 56..63
                                    6,  X,  X,  X,  X,  X,  X,  X,  // 64..71
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 72..79
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 80..87
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 88..95
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 96..103
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 104..111
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 112..119
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 120..127
                                    7,  X,  X,  X,  X,  X,  X,  X,  // 128..135
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 136..143
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 144..151
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 152..159
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 160..167
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 168..175
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 176..183
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 184..191
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 192..199
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 200..207
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 208..215
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 216..223
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 224..231
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 232..239
                                    X,  X,  X,  X,  X,  X,  X,  X,  // 240..247
                                    X,  X,  X,  X,  X,  X,  X,  X   // 248..255
                                   };
     union {
        UINT v;
        BYTE b[4];
     } value;
     value.v = n;
     if(value.b[0] != 0)
        return table[value.b[0]];
     if(value.b[1] != 0)
        return table[value.b[1]] + 8;
     if(value.b[2] != 0)
        return table[value.b[2]] + 16;
     if(value.b[3] != 0)
        return table[value.b[3]] + 24;
     return X;  // this is actually an error
#undef X
    } // TableTest

/****************************************************************************
*                                  NullTest
* Inputs:
*       UINT id: ignored
* Result: void
*       
* Effect: 
*       Does nothing.  The "Null" test
****************************************************************************/

__declspec(noinline) UINT NullTest(UINT id, UINT dummy)
    {
     return 0;
    } // NullTest


/****************************************************************************
*                                   LogTest
* Inputs:
*       UINT n: Value to compute
* Result: UINT
*       Log2(n)
* Notes: 
*       Does this by computing log(n)/log(2)
****************************************************************************/

__declspec(noinline) UINT LogTest(UINT n, UINT dummy)
    {
     return (UINT)(log((double)n)/log((double)2));
    } // LogTest

/****************************************************************************
*                                  Log10Test
* Inputs:
*       UINT n:
* Result: UINT
*       Log2(n)
* Effect: 
*       Computes this by log10(n)/log10(2)
****************************************************************************/

__declspec(noinline) UINT Log10Test(UINT n, UINT dummy)
    {
     return (UINT)(log10((double)n)/log10((double)2));
    } // Log10Test

/****************************************************************************
*                                  ShiftTest
* Inputs:
*       UINT n:
* Result: UINT
*       Log2(n)
* Effect: 
*       Computes the result by a >>1 loop
****************************************************************************/

__declspec(noinline) UINT ShiftTest(UINT n, UINT dummy)
    {
     UINT result;
     for(result = -1; n > 0; n >>= 1, result++) ;
     return result;
    } // ShiftTest

/****************************************************************************
*                                   BSFTest
* Inputs:
*       UINT n
* Result: UINT
*       Log2(n)
* Effect: 
*       COmputes Log2(n) using the BSF instruction
****************************************************************************/

__declspec(noinline) UINT BSFTest(UINT n, UINT dummy)
    {
     _asm bsf eax, n
    } // BSFTest

/****************************************************************************
*                                 FastBSFTest
* Inputs:
*       UINT n:
* Result: UINT
*       Log2(n)
* Notes: 
*       This uses the bsf instruction plus a fastcall linkage that should
*       also reduce call cost   
****************************************************************************/

__declspec(noinline) UINT __fastcall FastBSFTest(UINT n, UINT dummy)
    {
     _asm bsf eax, ecx
    } // FastBSFTest

/****************************************************************************
*                                    LOG2
* Inputs:
*       UINT n: Value to be computed
* Result: void
*       
* Effect: 
*       Computes the Log2(n) value with a variety of algorithms
****************************************************************************/

void LOG2(UINT n)
    {
     LARGE_INTEGER start;
     LARGE_INTEGER end;
     LARGE_INTEGER freq;
     UINT result;
     //----------------------------------------------------------------
     // class exp
     //----------------------------------------------------------------
     class EXP
         {
         public:
             EXP() { total = average = 0; }
             LONGLONG total;
             LONGLONG average;
             void SetAverage(UINT count) { average = total / count; }
         };
     //----------------------------------------------------------------
#define MAX_TESTS 10000

     QueryPerformanceFrequency(&freq);

     EXP null;
     //*****************************************************************************
     // NULL test
     //*****************************************************************************
     QueryPerformanceCounter(&start);
     for(int i = 0; i < MAX_TESTS; i++)
         { /* null test */
          OPTFENCE;
          result = NullTest(n, i);
         } /* null test */
     QueryPerformanceCounter(&end);
     null.total += end.QuadPart - start.QuadPart;
     null.SetAverage(MAX_TESTS);
     _tprintf(_T("\"NULL\", %I64d, %d, %I64d, %I64d, , \n"),
              null.total, MAX_TESTS, null.average, freq.QuadPart);

     //*****************************************************************************
     // log(n)/log(2)
     //*****************************************************************************
     QueryPerformanceCounter(&start);
     EXP logtest;
     for(int i = 0; i < MAX_TESTS; i++)
         { /* log2 test */
          OPTFENCE;
          result = LogTest(n, i);
         } /* log2 test */
     QueryPerformanceCounter(&end);
     logtest.total += end.QuadPart - start.QuadPart;

     logtest.SetAverage(MAX_TESTS);
     _tprintf(_T("\"Log(n)/Log(2)\", %I64d, %d, %I64d, %I64d, %u, %u\n"),
              logtest.total, MAX_TESTS, logtest.average, freq.QuadPart, n, result);

     //*****************************************************************************
     // Log10(n)/Log10(2)
     //*****************************************************************************
     EXP log10test;
     QueryPerformanceCounter(&start);
     for(int i = 0; i < MAX_TESTS; i++)
         { /* log10 test */
          OPTFENCE;
          result = Log10Test(n, i);
         }
     QueryPerformanceCounter(&end);
     log10test.total += end.QuadPart - start.QuadPart;
     log10test.SetAverage(MAX_TESTS);
     _tprintf(_T("\"Log10(n)/Log10(2)\", %I64d, %d, %I64d, %I64d, %u, %u\n"),
              log10test.total, MAX_TESTS, log10test.average, freq.QuadPart, n, result);

     //*****************************************************************************
     // table lookup
     //*****************************************************************************
     EXP tabletest;
     QueryPerformanceCounter(&start);
     for(int i = 0; i < MAX_TESTS; i++)
        { /* table test */
         OPTFENCE;
         result = TableTest(n, i);
        }
     QueryPerformanceCounter(&end);
     tabletest.total += end.QuadPart - start.QuadPart;
     tabletest.SetAverage(MAX_TESTS);
     _tprintf(_T("\"Lookup\", %I64d, %d, %I64d, %I64d, %u, %u\n"),
              tabletest.total, MAX_TESTS, tabletest.average, freq.QuadPart, n, result);

     //*****************************************************************************
     // >>1 loop count
     //*****************************************************************************

     EXP shifttest;
     QueryPerformanceCounter(&start);
     for(int i = 0; i < MAX_TESTS; i++)
         { /* shift test */
          OPTFENCE;
          result = ShiftTest(n, i);
         }
     QueryPerformanceCounter(&end);
     shifttest.total += end.QuadPart - start.QuadPart;
     shifttest.SetAverage(MAX_TESTS);
     _tprintf(_T("\">>1 loop\", %I64d, %d, %I64d, %I64d, %u, %u\n"),
              shifttest.total, MAX_TESTS, shifttest.average, freq.QuadPart, n, result);

     //*****************************************************************************
     // bsf instruction
     //*****************************************************************************
     EXP bsftest;
     QueryPerformanceCounter(&start);
     for(int i = 0; i < MAX_TESTS; i++)
         { /* bsf test */
          OPTFENCE;
          result = BSFTest(n, i);
         }
     QueryPerformanceCounter(&end);
     bsftest.total += end.QuadPart - start.QuadPart;
     bsftest.SetAverage(MAX_TESTS);
     _tprintf(_T("\"bsf instruction\", %I64d, %d, %I64d, %I64d, %u, %u\n"),
              bsftest.total, MAX_TESTS, bsftest.average, freq.QuadPart, n, result);

     //*****************************************************************************
     // bsf instruction (fast linkage)
     //*****************************************************************************
     EXP fastbsftest;
     QueryPerformanceCounter(&start);
     for(int i = 0; i < MAX_TESTS; i++)
        { /* bsf test */
         OPTFENCE;
         result = FastBSFTest(n, i);
        }
     QueryPerformanceCounter(&end);
     fastbsftest.total += end.QuadPart - start.QuadPart;
     fastbsftest.SetAverage(MAX_TESTS);
     _tprintf(_T("\"bsf instruction (fast)\", %I64d, %d, %I64d, %I64d, %u, %u\n"),
              fastbsftest.total, MAX_TESTS, fastbsftest.average, freq.QuadPart, n, result);
   }

/****************************************************************************
*                                   _tmain
* Inputs:
*       int argc: argument count
*       _TCHAR * argv[]: Array of arguments
* Result: int
*       Logically void, 0, always
* Effect: 
*       Driver code for the LOG2 test function
****************************************************************************/

int _tmain(int argc, _TCHAR* argv[])
   {
    for(UINT i = 0; i < 32; i++)
       LOG2(1<<i);
    return 0;
   }
